/*
 * src/instance.c
 *
 * Copyright (C) 2019 Ilja Kartašov <ik@lowenware.com>
 *
 * Project homepage: https://lowenware.com/aisl/
 *
 */

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/select.h>
#include <sys/time.h>

#include "list.h"
#include "str-utils.h"
#include "buffer.h"

#include "client.h"
#include "server.h"
#include "ssl.h"
//#include "globals.h"
#include "stream.h"
#include "instance.h"


struct callback
{
  void            * source;
  aisl_callback_t   f_ptr;
  aisl_event_t      event;
};

typedef struct callback * callback_t;


struct aisl
{
  list_t       server_spool;
  list_t       client_spool;
  list_t       callback_spool;
  #ifndef AISL_WITHOUT_SSL
  list_t       ssl_spool;
  #endif

  buffer_t     buffer;
  char       * last_error;

  size_t       iterator;

  uint32_t     accept_limit;
  uint32_t     silence_timeout;
};


#ifndef AISL_WITHOUT_SSL
static uint32_t m_instances = 0;
#endif


/* Initialization functions */

__attribute__ ((visibility ("default") ))
aisl_t
aisl_new( aisl_config_t config )
{
  aisl_t instance;

  #ifndef AISL_WITHOUT_SSL
  if ((m_instances++) == 0)
  {
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();
  }
  #endif

  if ( !(instance = calloc(1, sizeof(struct aisl))) )
    goto finally;

  if ( !(instance->server_spool = list_new(config->server_spool_size)) )
    goto release;

  if ( !(instance->client_spool = list_new(config->client_spool_size)) )
    goto release;

  if ( !(instance->callback_spool = list_new(config->callback_spool_size)) )
    goto release;

  if ( !(instance->buffer = buffer_new(config->initial_buffer_size)) )
    goto release;

  #ifndef AISL_WITHOUT_SSL
  if ( !(instance->ssl_spool = list_new(config->ssl_spool_size)) )
    goto release;
  #endif

  instance->accept_limit = config->client_accept_limit;
  instance->silence_timeout = config->client_silence_timeout;

  goto finally;

release:
  aisl_free(instance);
  instance = NULL;

finally:
  return instance;
}


__attribute__ ((visibility ("default") ))
void
aisl_free( aisl_t instance )
{
  if (instance->client_spool)
    list_free(instance->client_spool, (list_destructor_t) aisl_client_free );

  if (instance->server_spool)
    list_free(instance->server_spool, (list_destructor_t)  aisl_server_free );

  if (instance->callback_spool)
    list_free(instance->callback_spool, (list_destructor_t) free);

  if (instance->buffer)
    buffer_free(instance->buffer);

  #ifndef AISL_WITHOUT_SSL
  if (instance->ssl_spool)
    list_free(instance->ssl_spool, (list_destructor_t) ssl_free );
  #endif

  if (instance->last_error)
    free(instance->last_error);

  free(instance);

  #ifndef AISL_WITHOUT_SSL
  if ((--m_instances) == 0)
  {
    EVP_cleanup();
  }
  #endif
}


__attribute__ ((visibility ("default") ))
aisl_server_t
aisl_listen( aisl_t instance, const char * address, uint16_t port )
{
  aisl_server_t result;

  if ( (result = aisl_server_new(instance, address, port)) != NULL )
  {
    if (list_append(instance->server_spool, result) == LIST_NAN)
    {
      aisl_server_free(result);
    }
  }

  return result;
}


#ifndef AISL_WITHOUT_SSL

__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_set_ssl( aisl_t instance, const char * domain,
                               const char * key_file,
                               const char * crt_file )
{
  SSL_CTX * ssl_ctx = NULL;
  ssl_t ssl;
  size_t i;

  /* lookup for existing contexts */
  for (i=0; i<instance->ssl_spool->count; i++)
  {
    ssl = list_index(instance->ssl_spool, i);
    if (ssl->key_file && strcmp(ssl->key_file, key_file)==0 &&
        ssl->crt_file && strcmp(ssl->crt_file, crt_file)==0 )
    {
      ssl_ctx = ssl->ctx;
      break;
    }
  }

  if ((ssl = ssl_new(domain, key_file, crt_file, ssl_ctx)) != NULL)
  {
    if (list_append(instance->ssl_spool, ssl) != LIST_NAN)
      return AISL_SUCCESS;

    ssl_free(ssl);
  }

  return AISL_MALLOC_ERROR;
}


SSL_CTX *
aisl_get_ssl_ctx( aisl_t instance, const char * domain )
{
  size_t i;
  ssl_t ssl;

  if (domain)
  {
    for (i=0; i<instance->ssl_spool->count; i++)
    {
      ssl = list_index(instance->ssl_spool, i);
      if (strcmp(ssl->domain, domain) != 0)
        continue;

      return ssl->ctx;
    }
  }

  return NULL;
}


#endif

__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_set_callback( aisl_t           instance,
                   void           * source,
                   aisl_event_t     event,
                   aisl_callback_t  callback  )
{
  callback_t cb;

  if ( (cb = malloc(sizeof(struct callback))) != NULL )
  {
    cb->source = source;
    cb->event  = event;
    cb->f_ptr  = callback;

    if ( list_append(instance->callback_spool, cb) != LIST_NAN )
    {
      switch(event)
      {
        case AISL_EVENT_STREAM_OUTPUT:
          if (source)
            aisl_stream_set_chunked_output( (aisl_stream_t) source, true );
            //( (stream_t) source )->flags |= STREAM_FLAG_OUTPUT_CHUNKED;
          break;

        /*
         * case AISL_EVENT_STREAM_OPEN:
         *   instance->flags |= AISL_FLAG_HAS_STREAM_CALLBACKS;
         */
        default:
          break;
      }

      return AISL_SUCCESS;
    }
    free(cb);
  }
  return AISL_MALLOC_ERROR;
}


__attribute__ ((visibility ("default") ))
void
aisl_unset_callbacks_for( aisl_t instance, void * source )
{
  size_t i = instance->callback_spool->count;

  if (i)
  {
    for(i=i-1; i <= 0; i-- )
    {
      callback_t callback = list_index(instance->callback_spool, i);
      if ( callback->source == source )
      {
        free(callback);
        list_remove_index(instance->callback_spool, i);
      }
    }
  }
}


__attribute__ ((visibility ("default") ))
bool
aisl_raise( aisl_t instance, void * source, aisl_event_t event, ... )
{
  va_list vl;
  bool result;

  va_start(vl, event);
  result = aisl_raise_vl(instance, source, event, vl);
  va_end(vl);

  return result;
}


static bool
aisl_on_source_event( void * source, aisl_callback_t cb, va_list vl_args )
{
  aisl_on_source_event_t callback = (aisl_on_source_event_t) cb;

  (void) vl_args;

  return callback( (void *) source);
}


static bool
aisl_on_stream_open( void * source, aisl_callback_t cb, va_list vl_args )
{
  aisl_on_stream_open_t callback = (aisl_on_stream_open_t) cb;

  aisl_http_method_t method = va_arg(vl_args, aisl_http_method_t);
  char * path  = va_arg(vl_args, char *);
  char * query = va_arg(vl_args, char *);

  return callback( (aisl_stream_t) source, method, path, query);
}


static bool
aisl_on_stream_header( void * source, aisl_callback_t cb, va_list vl_args )
{
  aisl_on_stream_header_t callback = (aisl_on_stream_header_t) cb;

  char * key = va_arg(vl_args, char *);
  char * val = va_arg(vl_args, char *);

  return callback( (aisl_stream_t) source, key, val );
}


static bool
aisl_on_stream_input( void * source, aisl_callback_t cb, va_list vl_args )
{
  aisl_on_stream_input_t callback = (aisl_on_stream_input_t) cb;

  size_t len  = va_arg(vl_args, size_t);
  char * data = va_arg(vl_args, char *);

  return callback( (aisl_stream_t) source, data, len );
}


static bool
aisl_on_stream_output( void * source, aisl_callback_t cb, va_list vl_args )
{
  aisl_on_stream_output_t callback = (aisl_on_stream_output_t) cb;

  size_t buffer_size = va_arg(vl_args, size_t);

  return callback( (aisl_stream_t) source, buffer_size );
}


__attribute__ ((visibility ("default") ))
bool
aisl_raise_vl( aisl_t            instance,
               void            * source,
               aisl_event_t      event,
               va_list           vl_args )
{
  va_list vl_copy;
  size_t i;

  callback_t cb;
  bool result = false,
       stop_propg = false;

  i = instance->callback_spool->count;

  while(i--)
  {
    cb = list_index(instance->callback_spool, i);

    if (cb->event == event && (source == cb->source || cb->source == NULL))
    {
      va_copy(vl_copy, vl_args);
      switch(event)
      {
        /* server events */
        case AISL_EVENT_SERVER_OPEN:
        case AISL_EVENT_SERVER_ERROR:
        case AISL_EVENT_CLIENT_CONNECT:
        case AISL_EVENT_CLIENT_DISCONNECT:
        case AISL_EVENT_CLIENT_TIMEOUT:
        case AISL_EVENT_STREAM_REQUEST:
        case AISL_EVENT_STREAM_CLOSE:
        case AISL_EVENT_STREAM_ERROR:
          stop_propg = aisl_on_source_event(source, cb->f_ptr, vl_copy);
          break;

        case AISL_EVENT_STREAM_OPEN:
          stop_propg = aisl_on_stream_open(source, cb->f_ptr, vl_copy);
          break;

        case AISL_EVENT_STREAM_HEADER:
          stop_propg = aisl_on_stream_header(source, cb->f_ptr, vl_copy);
          break;

        case AISL_EVENT_STREAM_INPUT:
          stop_propg = aisl_on_stream_input(source, cb->f_ptr, vl_copy);
          break;

        case AISL_EVENT_STREAM_OUTPUT:
          stop_propg = aisl_on_stream_output(source, cb->f_ptr, vl_copy);
          break;

        default:
           stop_propg = ((aisl_on_custom_event_t) cb->f_ptr)(source, vl_copy);
      }

      va_end(vl_copy);

      result = true;

      if (stop_propg)
        break;
    }
  }

  return result;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_run_cycle( aisl_t instance )
{
  aisl_status_t result = AISL_IDLE;

  size_t i = instance->iterator,
         max = instance->server_spool->count + instance->client_spool->count;

  while ( result == AISL_IDLE && max != 0)
  {
    aisl_client_t cli;

    if (i < instance->server_spool->count)
    {
      if (instance->client_spool->count < instance->accept_limit)
      {
        aisl_server_t srv;
        SSL_CTX * ssl_ctx;

        srv = (aisl_server_t) list_index(instance->server_spool, i);

        ssl_ctx = (aisl_server_get_ssl(srv)) ?
                    aisl_get_ssl_ctx(instance, NULL) :
                    NULL;

        result = aisl_server_touch(srv, &cli, ssl_ctx);

        if (result == AISL_SUCCESS)
        {
          if ( list_append(instance->client_spool, cli) != LIST_NAN )
          {
            aisl_raise(instance, cli, AISL_EVENT_CLIENT_CONNECT);
          }
          else
          {
            aisl_client_free(cli);
            /* todo: save error message here */
          }
        }
      }
      else
      {
        i = instance->server_spool->count;
        continue;
      }
    }
    else
    {
      size_t j = i - instance->server_spool->count;

      if (j < instance->client_spool->count)
      {
        aisl_client_t c = list_index(instance->client_spool, j);

        result = aisl_client_touch(c);

        if (aisl_client_is_timed_out( c, instance->silence_timeout ) )
          aisl_raise( instance, c, AISL_EVENT_CLIENT_TIMEOUT );

        if ( !aisl_client_is_online(c) )
        {
          aisl_client_free( c );
          list_remove_index(instance->client_spool, j);
        }
      }
      else
      {
        i = 0;
        continue;
      }
    }

    max--;
    i++;
  }

  instance->iterator = i;

  return result;
}


__attribute__ ((visibility ("default") ))
const char *
aisl_get_error( aisl_t instance )
{
  return instance->last_error;
}


void
aisl_set_error( aisl_t instance, const char * err_msg )
{
  if (instance->last_error)
    free(instance->last_error);

  instance->last_error = str_copy(err_msg);
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_sleep( aisl_t instance, uint32_t usec )
{
  int maxfd=0,
      sd;
  size_t i;
  struct timeval timeout = {0,usec};

  memset(&timeout, 0, sizeof(struct timeval));
  timeout.tv_usec = usec;

  fd_set  fs;
  FD_ZERO (&fs);

  for (i=0; i<instance->server_spool->count; i++)
  {
    aisl_server_t s = list_index(instance->server_spool, i);

    sd = aisl_server_get_socket(s);

    if (sd != -1)
    {
      FD_SET(sd, &fs);
      if (sd > maxfd) maxfd = sd;
    }
  }


  for (i=0; i<instance->client_spool->count; i++)
  {
    aisl_client_t c = list_index(instance->client_spool, i);
    sd = aisl_client_get_socket(c);
    if (sd != -1)
    {
      FD_SET(sd, &fs);
      if (sd > maxfd) maxfd = sd;
    }
  }

  switch ( select(maxfd+1, &fs, NULL, NULL, &timeout) )
  {
    case -1:
      return AISL_SYSCALL_ERROR;

    case 0:
      return AISL_IDLE;

    default:
      return AISL_SUCCESS;
  }
}

