#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>

#ifdef __APPLE__

#include <sys/ioctl.h>

#endif

#include <openssl/ssl.h>

#include "str-utils.h"
#include "instance.h"
#include "client.h"
#include "server.h"


/**
 * @brief HTTP(s) server data structure represented by #aisl_server_t pointer.
 */
struct aisl_server
{
  struct sockaddr_in   address;  /**< TCP server address to listen to. */
  aisl_t               instance; /**< Associated AISL instance pointer. */
  int                  fd;       /**< System socket descriptor. */
  bool                 ssl;      /**< SSL enabled/disabled flag. */
};


/**
 * @brief Creates TCP server socket, binds to address and starts to listen.
 * @param server a pointer to #aisl_server_t instance.
 * @return #aisl_status_t code.
 */
static aisl_status_t
aisl_server_open(aisl_server_t server)
{
  int fd, s_opt  = 1;

  fd = socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);

  if (fd != -1)
  {
    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void *)&s_opt, sizeof(int));

    #ifdef __APPLE__
    ioctl(fd, FIONBIO, (char *)&s_opt);
    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    #endif

    s_opt = sizeof(struct sockaddr_in);

    if (bind(fd, (struct sockaddr *) &server->address, s_opt)==0)
    {
      if (listen(fd, SOMAXCONN) == 0)
      {
        server->fd = fd;
        return AISL_SUCCESS;
      }
    }

    close(fd);
  }

  return AISL_SYSCALL_ERROR;
}

/**
 * @brief Tries to accept a new client.
 * @param server a pointer to #aisl_server_t instance.
 * @param p_client a pointer to store #aisl_client_t instance pointer.
 * @param ssl_ctx a pointer to SSL context or NULL if connection is unsecure.
 * @return #aisl_status_t code.
 */
static aisl_status_t
aisl_server_accept( aisl_server_t   server,
                    aisl_client_t * p_client,
                    SSL_CTX       * ssl_ctx )
{
  int                fd;
  struct sockaddr_in addr;
  socklen_t          len = sizeof(struct sockaddr_in);

  *p_client = NULL;

  fd = accept(server->fd, (struct sockaddr *) &addr, &len);

  if (fd != -1)
  {
    int flags;

    if ((flags = fcntl(fd, F_GETFL, 0)) != -1)
    {
      flags |= O_NONBLOCK;

      if (fcntl(fd, F_SETFL, flags) == 0)
      {
        return (!(*p_client = aisl_client_new(server, fd, &addr, ssl_ctx))) ?
           AISL_MALLOC_ERROR : AISL_SUCCESS;
      }
    }
    close(fd);
  }
  else if (errno == EWOULDBLOCK)
    return AISL_IDLE;

  aisl_raise( server->instance, server, AISL_EVENT_SERVER_ERROR );

  return AISL_SYSCALL_ERROR;
}

/* Library Level ------------------------------------------------------------ */

aisl_status_t
aisl_server_touch( aisl_server_t   server,
                   aisl_client_t * p_client,
                   SSL_CTX       * ssl_ctx )
{
  aisl_status_t result;

  if (server->fd == -1)
  {
    aisl_event_t event;

    result = aisl_server_open(server);
    event  = (result == AISL_SUCCESS) ? AISL_EVENT_SERVER_OPEN
                                      : AISL_EVENT_SERVER_ERROR;

    aisl_raise(server->instance, server, event);
  }
  else
    result = aisl_server_accept(server, p_client, ssl_ctx);

  return result;
}

/* API Level ---------------------------------------------------------------- */

aisl_server_t
aisl_server_new(aisl_t instance, const char * host, uint16_t port)
{
  aisl_server_t server;

  if ( (server = calloc(1, sizeof(struct aisl_server))) != NULL )
  {
    server->instance = instance;
    server->fd = -1;
    server->address.sin_family = AF_INET;
    server->address.sin_addr.s_addr = inet_addr(host);
    server->address.sin_port = htons(port);
  }

  return server;
}


void
aisl_server_free(aisl_server_t server)
{
  if (server)
  {
    if ( server->fd != -1)
    {
      close(server->fd);
      server->fd=-1;
    }

    free(server);
  }
}


aisl_t
aisl_server_get_instance( aisl_server_t server )
{
  return server->instance;
}


void
aisl_server_get_address( aisl_server_t server, struct sockaddr_in * address)
{
  memcpy(address, &server->address, sizeof(struct sockaddr_in));
}


#ifndef AISL_WITHOUT_SSL

void
aisl_server_set_ssl( aisl_server_t server, bool value )
{
  server->ssl = value;
}


bool
aisl_server_get_ssl( aisl_server_t server )
{
  return server->ssl;
}

#endif
