#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "buffer.h"
#include "stream.h"

#define FLAG_OUTPUT_READY   (1<<0)
#define FLAG_OUTPUT_CHUNKED (1<<1)

struct aisl_stream
{
  aisl_client_t          client;

  struct buffer          buffer;

  const char            *content_type;

  size_t                 content_length;
  int32_t                content_offset;
  int                    id;
  int                    flags;

  aisl_http_response_t   response;
  aisl_stream_state_t    state;

  bool                   c_length_unknown;

};
static void
pair_free( pair_t stream )
{
  if (!stream) return;

  if(stream->key) free(stream->key);
  if(stream->value) free(stream->value);
  free(stream);
}
/* -------------------------------------------------------------------------- */

aisl_stream_t
stream_new(struct sockaddr_in  *client, int id, stream_state_t state)
{
  aisl_stream_t stream = malloc(sizeof(struct aisl_stream));

  if (stream)
  {
    /* public data */
    ASTREAM(stream)->client         = client;
    ASTREAM(stream)->host           = NULL;
    ASTREAM(stream)->path           = NULL;
    ASTREAM(stream)->query          = NULL;
    ASTREAM(stream)->scheme         = NULL;
    ASTREAM(stream)->u_ptr          = NULL;
    ASTREAM(stream)->request_method = AISL_HTTP_GET;

    /* private data */
    stream->headers                 = NULL; /* request headers */
    stream->buffer                  = buffer_new(0);

    stream->c_type                  = NULL;

    stream->response                = AISL_HTTP_OK;
    stream->state                   = STREAM_REQUEST_METHOD;
    stream->c_length                = 0;
    stream->c_offset                = 0; /* headers length */
    stream->id                      = id;
    stream->c_length_unknown        = true;
    stream->flags                   = 0;
  }
  return stream;
}

/*
aisl_stream_t
stream_reset(aisl_stream_t stream)
{
  if (ASTREAM(stream)->path)
  {
    free( (char*) ASTREAM(stream)->path);
    ASTREAM(stream)->path = NULL;
  }

  if (ASTREAM(stream)->query)
  {
    free( (char*) ASTREAM(stream)->query);
    ASTREAM(stream)->query = NULL;
  }

  ASTREAM(stream)->u_ptr          = NULL;
  ASTREAM(stream)->request_method = AISL_HTTP_GET;

  if (stream->headers)
  {
    list_free(stream->headers, (list_destructor_t) pair_free);
    stream->headers = NULL;
  }

  stream->c_type                  = NULL;
  stream->response                = AISL_HTTP_OK;
  stream->state                   = STREAM_REQUEST_METHOD;
  stream->c_length                = 0;
  stream->c_offset                = 0; / * headers length * /
  stream->c_length_unknown        = true;
  stream->flags                   = 0;

  return stream;
}
*/
/* -------------------------------------------------------------------------- */

void
stream_free(aisl_stream_t stream)
{
  if (stream->buffer)  buffer_free(stream->buffer);
  if (stream->headers) list_free(stream->headers, (list_destructor_t) pair_free);

  if (ASTREAM(stream)->path)  free( (char*) ASTREAM(stream)->path);
  if (ASTREAM(stream)->query) free( (char*) ASTREAM(stream)->query);

  aisl_handle_t hd =  ((client_t) ASTREAM(stream)->client)->server->owner;
  aisl_raise_event(hd, stream, AISL_STREAM_CLOSE);
  ASTREAM(stream)->u_ptr = NULL;
  aisl_remove_listeners_for(hd, stream);
  free(stream);
}

/* -------------------------------------------------------------------------- */

int
stream_write(aisl_stream_t stream, const char * data, uint32_t d_len)
{
  return buffer_add( stream->buffer, data, d_len);
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void
aisl_cancel(aisl_aisl_stream_t s)
{
  client_close( (client_t) s->client );
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
bool
aisl_is_secure(aisl_aisl_stream_t s)
{
  client_t cli = (client_t) s->client;

  return (cli->ssl) ? true : false;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void *
aisl_get_context(aisl_aisl_stream_t s)
{
  return s->u_ptr;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void
aisl_set_context(aisl_aisl_stream_t s, void * u_ptr)
{
  s->u_ptr = u_ptr;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
aisl_client_t
aisl_get_client(aisl_aisl_stream_t s)
{
  return s->client;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
aisl_server_t
aisl_get_server(aisl_aisl_stream_t s)
{
  return  (aisl_server_t) (((client_t) s->client)->server);
}


/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
aisl_http_version_t
aisl_get_http_version(aisl_aisl_stream_t s)
{
  client_t cli = (client_t) s->client;

  return cli->protocol;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void
aisl_reject(aisl_aisl_stream_t s)
{
  client_t cli = (client_t) s->client;

  client_close( cli );
}

/* -------------------------------------------------------------------------- */



static char *
get_response_begin(aisl_stream_t stream)
{
  char * r;
  client_t cli = CLIENT(ASTREAM(stream)->client);

  r = str_printf(
        "%s %d %s\r\n"
        "Server: AISL\r\n"
        "Connection: %s\r\n\r\n",
        aisl_http_version_to_string(cli->protocol),
        stream->response,
        aisl_http_response_to_string(stream->response),
        ((cli->flags & CLIENT_FLAG_KEEPALIVE) ? "keep-alive" : "close")
      );

  return r;
}

__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_response(aisl_aisl_stream_t stream, aisl_http_response_t  status_code,
                                    const char           *content_type,
                                    uint32_t              content_length)
{
  char  * pch;
  int     l;


  /* check if those headers were already sent */
  if (STREAM(stream)->state > STREAM_REQUEST_READY) return AISL_IDLE;

  STREAM(stream)->response = status_code;
  STREAM(stream)->c_type   = content_type;
  STREAM(stream)->c_length = content_length;

  if ( !(pch = get_response_begin(STREAM(stream))) )
    return AISL_MALLOC_ERROR;

  l = strlen(pch);
  STREAM(stream)->c_offset = l-2;

  buffer_clear(STREAM(stream)->buffer, content_length);

  l = buffer_add( STREAM(stream)->buffer, pch, l );
  free(pch);

  if (l == BUFFER_EOB) return AISL_MALLOC_ERROR;

  if (content_length)
  {
    if (!aisl_header_printf( stream, "Content-Length", "%u", content_length ))
      return AISL_MALLOC_ERROR;
  }

  if (content_type)
  {
    if (!aisl_header( stream, "Content-Type", content_type ))
      return AISL_MALLOC_ERROR;
  }

  STREAM(stream)->state = STREAM_RESPONSE_HEADER;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_flush(aisl_aisl_stream_t stream)
{
  aisl_stream_t s = STREAM(stream); 
  if ( ! s->c_length )
  {
    s->c_length = s->buffer->size - s->c_offset - 2;
    if (!aisl_header_printf(stream, "Content-Length", "%u", s->c_length))
      return AISL_MALLOC_ERROR;
  }

  /*
  fprintf(stdout, "(%lu bytes)------->\n", STREAM(stream)->buffer->size);
  fwrite(STREAM(stream)->buffer->data, 1, STREAM(stream)->buffer->size, stdout);
  fprintf(stdout, "<------\n");
  */
  s->state = STREAM_RESPONSE_READY;

  s->flags |= STREAM_FLAG_OUTPUT_READY;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
int
aisl_header(aisl_aisl_stream_t stream, const char *key, const char *value)
{
  int ret;
  char * pch;

  if ( (pch = str_printf("%s: %s\r\n", key, value)) != NULL )
  {
    ret = strlen(pch);
    if ( buffer_insert(
           STREAM(stream)->buffer,
           STREAM(stream)->c_offset,
           pch,
           ret
         ) == BUFFER_EOB )
    {
      ret = -1;
    }
    else
      STREAM(stream)->c_offset += ret;

    free(pch);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_printf(aisl_aisl_stream_t stream, const char *key,
                                         const char *f_value, ...)
{
  int ret;

  va_list arg;
  va_start(arg, f_value);

  ret = aisl_header_vprintf(stream, key, f_value, arg);

  va_end(arg);

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_vprintf(aisl_aisl_stream_t stream, const char *key,
                                          const char *format,
                                          va_list     args)
{
  int ret;
  char * value;

  if ( (value = str_vprintf(format, args)) != NULL )
  {
    ret = aisl_header( stream, key, value );
    free(value);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_printf(aisl_aisl_stream_t stream, const char *format, ...)
{
  va_list arg;
  va_start(arg, format);

  int result = aisl_vprintf(stream, format, arg);

  va_end(arg);

  /* No need to update length there, because vprintf do that
   * 
   * if (STREAM(stream)->c_length_unknown)
    STREAM(stream)->c_length += result;
    */


  return result;

}


__attribute__ ((visibility ("default") ))
int
aisl_vprintf(aisl_aisl_stream_t stream, const char *format, va_list args)
{
  int    result;
  char * r;

  if ( (r = str_vprintf(format, args)) != NULL)
  {
    result = strlen(r);
    if (buffer_add(STREAM(stream)->buffer, r, result) == BUFFER_EOB)
      result = -1;

    free(r);
  }
  else
    result = -1;

  return result;
}


__attribute__ ((visibility ("default") ))
int
aisl_write(aisl_aisl_stream_t stream, const char *data, int d_len)
{
  if (d_len < 0)
    d_len = strlen(data);

  if (buffer_add(STREAM(stream)->buffer, data, d_len) == BUFFER_EOB)
    d_len = -1;

  return d_len;
}


__attribute__ ((visibility ("default") ))
int
aisl_puts(const char *str, aisl_aisl_stream_t stream)
{
  return aisl_write( stream, str, strlen(str));
}


